import kotlinx.coroutines.*
import kotlinx.coroutines.channels.*
import kotlin.coroutines.CoroutineContext

fun main() {

  val firstChannel = GlobalScope
      .produce(capacity = Channel.UNLIMITED) {
        offer("A")
        offer("B")

        Thread.sleep(3000)
        offer("D")
      }

  val secondChannel = GlobalScope
      .produce(capacity = Channel.UNLIMITED) {
        offer(4)
        offer(1)

        Thread.sleep(5000)
        offer(100)
      }

  val result = GlobalScope.zip(firstChannel, secondChannel) { first, second ->
    "$first$second"
  }

  GlobalScope.launch {
    result.consumeEach { println(it) }
  }

  Thread.sleep(1000)

  while (true) {
  }
}

inline fun <T, R, U> CoroutineScope.zip(
    first: ReceiveChannel<T>,
    second: ReceiveChannel<R>,
    crossinline mapper: (T, R) -> U)
    : ReceiveChannel<U> {
  return produce(capacity = Channel.UNLIMITED) {
    var firstValue: T? = null
    var secondValue: R? = null

    while (isActive) {
      if (firstValue == null) {
        firstValue = first.poll()
      }

      if (secondValue == null) {
        secondValue = second.poll()
      }

      if (firstValue != null && secondValue != null) {
        val combinedValue = mapper(firstValue, secondValue)

        offer(combinedValue)

        firstValue = null
        secondValue = null
      }
    }
  }
}

inline fun <T> mergeAndConsume(
    consumerScope: CoroutineScope,
    processorScope: CoroutineScope,
    first: ReceiveChannel<T>,
    second: ReceiveChannel<T>,
    crossinline consumer: (T) -> Unit): () -> Unit {

  val actor = consumerScope.actor<T>(capacity = Channel.UNLIMITED) {
    consumeEach(consumer)
  }

  processorScope.launch {
    first.consumeEach { actor.offer(it) }
  }

  processorScope.launch {
    second.consumeEach { actor.offer(it) }
  }

  return {
    first.cancel()
    second.cancel()
    actor.close()
  }
}

object IntervalEvent

fun CoroutineScope.interval(
    coroutineContext: CoroutineContext = Dispatchers.Default,
    delayInterval: Long = 1000): ReceiveChannel<IntervalEvent> =
    produce(coroutineContext) {
      while (isActive) {
        offer(IntervalEvent)
        delay(delayInterval)
      }
    }

inline fun ReceiveChannel<IntervalEvent>.subscribe(coroutineScope: CoroutineScope,
                                                   crossinline action: () -> Unit) {
  coroutineScope.launch {
    consumeEach { action() }
  }
}